=== modified file 'plugins.d/mandos-client.c'
--- plugins.d/mandos-client.c	2011-11-11 09:07:09 +0000
+++ plugins.d/mandos-client.c	2011-11-12 17:43:12 +0000
@@ -53,7 +53,7 @@
 				   sockaddr_in6, PF_INET6,
 				   SOCK_STREAM, uid_t, gid_t, open(),
 				   opendir(), DIR */
-#include <sys/stat.h>		/* open() */
+#include <sys/stat.h>		/* open(), S_ISREG */
 #include <sys/socket.h>		/* socket(), struct sockaddr_in6,
 				   inet_pton(), connect() */
 #include <fcntl.h>		/* open() */
@@ -85,6 +85,8 @@
 				   raise() */
 #include <sysexits.h>		/* EX_OSERR, EX_USAGE, EX_UNAVAILABLE,
 				   EX_NOHOST, EX_IOERR, EX_PROTOCOL */
+#include <sys/wait.h>		/* waitpid(), WIFEXITED(),
+				   WEXITSTATUS(), WTERMSIG() */
 
 #ifdef __linux__
 #include <sys/klog.h> 		/* klogctl() */
@@ -123,6 +125,7 @@
 #define PATHDIR "/conf/conf.d/mandos"
 #define SECKEY "seckey.txt"
 #define PUBKEY "pubkey.txt"
+#define HOOKDIR "/lib/mandos/network-hooks.d"
 
 bool debug = false;
 static const char mandos_protocol_version[] = "1";
@@ -130,6 +133,7 @@
 const char *argp_program_bug_address = "<mandos@recompile.se>";
 static const char sys_class_net[] = "/sys/class/net";
 char *connect_to = NULL;
+const char *hookdir = HOOKDIR;
 
 /* Doubly linked list that need to be circularly linked when used */
 typedef struct server{
@@ -250,14 +254,16 @@
     
     rc = gpgme_data_new_from_fd(&pgp_data, fd);
     if(rc != GPG_ERR_NO_ERROR){
-      fprintf(stderr, "bad gpgme_data_new_from_fd: %s: %s\n",
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "bad gpgme_data_new_from_fd: %s: %s\n",
 	      gpgme_strsource(rc), gpgme_strerror(rc));
       return false;
     }
     
     rc = gpgme_op_import(mc.ctx, pgp_data);
     if(rc != GPG_ERR_NO_ERROR){
-      fprintf(stderr, "bad gpgme_op_import: %s: %s\n",
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "bad gpgme_op_import: %s: %s\n",
 	      gpgme_strsource(rc), gpgme_strerror(rc));
       return false;
     }
@@ -271,14 +277,16 @@
   }
   
   if(debug){
-    fprintf(stderr, "Initializing GPGME\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Initializing GPGME\n");
   }
   
   /* Init GPGME */
   gpgme_check_version(NULL);
   rc = gpgme_engine_check_version(GPGME_PROTOCOL_OpenPGP);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "bad gpgme_engine_check_version: %s: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "bad gpgme_engine_check_version: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
     return false;
   }
@@ -286,7 +294,8 @@
   /* Set GPGME home directory for the OpenPGP engine only */
   rc = gpgme_get_engine_info(&engine_info);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "bad gpgme_get_engine_info: %s: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "bad gpgme_get_engine_info: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
     return false;
   }
@@ -299,15 +308,17 @@
     engine_info = engine_info->next;
   }
   if(engine_info == NULL){
-    fprintf(stderr, "Could not set GPGME home dir to %s\n", tempdir);
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Could not set GPGME home dir to %s\n", tempdir);
     return false;
   }
   
   /* Create new GPGME "context" */
   rc = gpgme_new(&(mc.ctx));
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "bad gpgme_new: %s: %s\n",
-	    gpgme_strsource(rc), gpgme_strerror(rc));
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "bad gpgme_new: %s: %s\n", gpgme_strsource(rc),
+	    gpgme_strerror(rc));
     return false;
   }
   
@@ -332,14 +343,16 @@
   ssize_t plaintext_length = 0;
   
   if(debug){
-    fprintf(stderr, "Trying to decrypt OpenPGP data\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Trying to decrypt OpenPGP data\n");
   }
   
   /* Create new GPGME data buffer from memory cryptotext */
   rc = gpgme_data_new_from_mem(&dh_crypto, cryptotext, crypto_size,
 			       0);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "bad gpgme_data_new_from_mem: %s: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "bad gpgme_data_new_from_mem: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
     return -1;
   }
@@ -347,7 +360,8 @@
   /* Create new empty GPGME data buffer for the plaintext */
   rc = gpgme_data_new(&dh_plain);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "bad gpgme_data_new: %s: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "bad gpgme_data_new: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
     gpgme_data_release(dh_crypto);
     return -1;
@@ -357,29 +371,37 @@
      data buffer */
   rc = gpgme_op_decrypt(mc.ctx, dh_crypto, dh_plain);
   if(rc != GPG_ERR_NO_ERROR){
-    fprintf(stderr, "bad gpgme_op_decrypt: %s: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "bad gpgme_op_decrypt: %s: %s\n",
 	    gpgme_strsource(rc), gpgme_strerror(rc));
     plaintext_length = -1;
     if(debug){
       gpgme_decrypt_result_t result;
       result = gpgme_op_decrypt_result(mc.ctx);
       if(result == NULL){
-	fprintf(stderr, "gpgme_op_decrypt_result failed\n");
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"gpgme_op_decrypt_result failed\n");
       } else {
-	fprintf(stderr, "Unsupported algorithm: %s\n",
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Unsupported algorithm: %s\n",
 		result->unsupported_algorithm);
-	fprintf(stderr, "Wrong key usage: %u\n",
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Wrong key usage: %u\n",
 		result->wrong_key_usage);
 	if(result->file_name != NULL){
-	  fprintf(stderr, "File name: %s\n", result->file_name);
+	  fprintf(stderr, "Mandos plugin mandos-client: "
+		  "File name: %s\n", result->file_name);
 	}
 	gpgme_recipient_t recipient;
 	recipient = result->recipients;
 	while(recipient != NULL){
-	  fprintf(stderr, "Public key algorithm: %s\n",
+	  fprintf(stderr, "Mandos plugin mandos-client: "
+		  "Public key algorithm: %s\n",
 		  gpgme_pubkey_algo_name(recipient->pubkey_algo));
-	  fprintf(stderr, "Key ID: %s\n", recipient->keyid);
-	  fprintf(stderr, "Secret key available: %s\n",
+	  fprintf(stderr, "Mandos plugin mandos-client: "
+		  "Key ID: %s\n", recipient->keyid);
+	  fprintf(stderr, "Mandos plugin mandos-client: "
+		  "Secret key available: %s\n",
 		  recipient->status == GPG_ERR_NO_SECKEY
 		  ? "No" : "Yes");
 	  recipient = recipient->next;
@@ -390,7 +412,8 @@
   }
   
   if(debug){
-    fprintf(stderr, "Decryption of OpenPGP data succeeded\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Decryption of OpenPGP data succeeded\n");
   }
   
   /* Seek back to the beginning of the GPGME plaintext data buffer */
@@ -427,7 +450,8 @@
   }
   
   if(debug){
-    fprintf(stderr, "Decrypted password is: ");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Decrypted password is: ");
     for(ssize_t i = 0; i < plaintext_length; i++){
       fprintf(stderr, "%02hhX ", (*plaintext)[i]);
     }
@@ -455,7 +479,7 @@
 /* GnuTLS log function callback */
 static void debuggnutls(__attribute__((unused)) int level,
 			const char* string){
-  fprintf(stderr, "GnuTLS: %s", string);
+  fprintf(stderr, "Mandos plugin mandos-client: GnuTLS: %s", string);
 }
 
 static int init_gnutls_global(const char *pubkeyfilename,
@@ -463,13 +487,14 @@
   int ret;
   
   if(debug){
-    fprintf(stderr, "Initializing GnuTLS\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Initializing GnuTLS\n");
   }
   
   ret = gnutls_global_init();
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "GnuTLS global_init: %s\n",
-	    safer_gnutls_strerror(ret));
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "GnuTLS global_init: %s\n", safer_gnutls_strerror(ret));
     return -1;
   }
   
@@ -484,14 +509,15 @@
   /* OpenPGP credentials */
   ret = gnutls_certificate_allocate_credentials(&mc.cred);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "GnuTLS memory error: %s\n",
-	    safer_gnutls_strerror(ret));
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "GnuTLS memory error: %s\n", safer_gnutls_strerror(ret));
     gnutls_global_deinit();
     return -1;
   }
   
   if(debug){
-    fprintf(stderr, "Attempting to use OpenPGP public key %s and"
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Attempting to use OpenPGP public key %s and"
 	    " secret key %s as GnuTLS credentials\n", pubkeyfilename,
 	    seckeyfilename);
   }
@@ -501,23 +527,26 @@
      GNUTLS_OPENPGP_FMT_BASE64);
   if(ret != GNUTLS_E_SUCCESS){
     fprintf(stderr,
+	    "Mandos plugin mandos-client: "
 	    "Error[%d] while reading the OpenPGP key pair ('%s',"
 	    " '%s')\n", ret, pubkeyfilename, seckeyfilename);
-    fprintf(stderr, "The GnuTLS error is: %s\n",
-	    safer_gnutls_strerror(ret));
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "The GnuTLS error is: %s\n", safer_gnutls_strerror(ret));
     goto globalfail;
   }
   
   /* GnuTLS server initialization */
   ret = gnutls_dh_params_init(&mc.dh_params);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Error in GnuTLS DH parameter initialization:"
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Error in GnuTLS DH parameter initialization:"
 	    " %s\n", safer_gnutls_strerror(ret));
     goto globalfail;
   }
   ret = gnutls_dh_params_generate2(mc.dh_params, mc.dh_bits);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Error in GnuTLS prime generation: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Error in GnuTLS prime generation: %s\n",
 	    safer_gnutls_strerror(ret));
     goto globalfail;
   }
@@ -544,7 +573,8 @@
     }
   } while(ret == GNUTLS_E_INTERRUPTED or ret == GNUTLS_E_AGAIN);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Error in GnuTLS session initialization: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Error in GnuTLS session initialization: %s\n",
 	    safer_gnutls_strerror(ret));
   }
   
@@ -558,9 +588,10 @@
       }
     } while(ret == GNUTLS_E_INTERRUPTED or ret == GNUTLS_E_AGAIN);
     if(ret != GNUTLS_E_SUCCESS){
-      fprintf(stderr, "Syntax error at: %s\n", err);
-      fprintf(stderr, "GnuTLS error: %s\n",
-	      safer_gnutls_strerror(ret));
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Syntax error at: %s\n", err);
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "GnuTLS error: %s\n", safer_gnutls_strerror(ret));
       gnutls_deinit(*session);
       return -1;
     }
@@ -575,7 +606,8 @@
     }
   } while(ret == GNUTLS_E_INTERRUPTED or ret == GNUTLS_E_AGAIN);
   if(ret != GNUTLS_E_SUCCESS){
-    fprintf(stderr, "Error setting GnuTLS credentials: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Error setting GnuTLS credentials: %s\n",
 	    safer_gnutls_strerror(ret));
     gnutls_deinit(*session);
     return -1;
@@ -627,7 +659,8 @@
     pf = PF_INET;
     break;
   default:
-    fprintf(stderr, "Bad address family: %d\n", af);
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Bad address family: %d\n", af);
     errno = EINVAL;
     return -1;
   }
@@ -638,7 +671,8 @@
   }
   
   if(debug){
-    fprintf(stderr, "Setting up a TCP connection to %s, port %" PRIu16
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Setting up a TCP connection to %s, port %" PRIu16
 	    "\n", ip, port);
   }
   
@@ -671,7 +705,8 @@
   }
   if(ret == 0){
     int e = errno;
-    fprintf(stderr, "Bad address: %s\n", ip);
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Bad address: %s\n", ip);
     errno = e;
     goto mandos_end;
   }
@@ -684,7 +719,8 @@
        (&to.in6.sin6_addr)){ /* -Wstrict-aliasing=2 or lower and
 			      -Wunreachable-code*/
       if(if_index == AVAHI_IF_UNSPEC){
-	fprintf(stderr, "An IPv6 link-local address is incomplete"
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"An IPv6 link-local address is incomplete"
 		" without a network interface\n");
 	errno = EINVAL;
 	goto mandos_end;
@@ -709,12 +745,13 @@
       if(if_indextoname((unsigned int)if_index, interface) == NULL){
 	perror_plus("if_indextoname");
       } else {
-	fprintf(stderr, "Connection to: %s%%%s, port %" PRIu16 "\n",
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Connection to: %s%%%s, port %" PRIu16 "\n",
 		ip, interface, port);
       }
     } else {
-      fprintf(stderr, "Connection to: %s, port %" PRIu16 "\n", ip,
-	      port);
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Connection to: %s, port %" PRIu16 "\n", ip, port);
     }
     char addrstr[(INET_ADDRSTRLEN > INET6_ADDRSTRLEN) ?
 		 INET_ADDRSTRLEN : INET6_ADDRSTRLEN] = "";
@@ -730,7 +767,8 @@
       perror_plus("inet_ntop");
     } else {
       if(strcmp(addrstr, ip) != 0){
-	fprintf(stderr, "Canonical address form: %s\n", addrstr);
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Canonical address form: %s\n", addrstr);
       }
     }
   }
@@ -790,7 +828,8 @@
   }
   
   if(debug){
-    fprintf(stderr, "Establishing TLS session with %s\n", ip);
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Establishing TLS session with %s\n", ip);
   }
   
   if(quit_now){
@@ -816,7 +855,8 @@
   
   if(ret != GNUTLS_E_SUCCESS){
     if(debug){
-      fprintf(stderr, "*** GnuTLS Handshake failed ***\n");
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "*** GnuTLS Handshake failed ***\n");
       gnutls_perror(ret);
     }
     errno = EPROTO;
@@ -826,8 +866,8 @@
   /* Read OpenPGP packet that contains the wanted password */
   
   if(debug){
-    fprintf(stderr, "Retrieving OpenPGP encrypted password from %s\n",
-	    ip);
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Retrieving OpenPGP encrypted password from %s\n", ip);
   }
   
   while(true){
@@ -871,14 +911,16 @@
 	  }
 	} while(ret == GNUTLS_E_AGAIN or ret == GNUTLS_E_INTERRUPTED);
 	if(ret < 0){
-	  fprintf(stderr, "*** GnuTLS Re-handshake failed ***\n");
+	  fprintf(stderr, "Mandos plugin mandos-client: "
+		  "*** GnuTLS Re-handshake failed ***\n");
 	  gnutls_perror(ret);
 	  errno = EPROTO;
 	  goto mandos_end;
 	}
 	break;
       default:
-	fprintf(stderr, "Unknown error while reading data from"
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Unknown error while reading data from"
 		" encrypted session with Mandos server\n");
 	gnutls_bye(session, GNUTLS_SHUT_RDWR);
 	errno = EIO;
@@ -890,7 +932,8 @@
   }
   
   if(debug){
-    fprintf(stderr, "Closing TLS session\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Closing TLS session\n");
   }
   
   if(quit_now){
@@ -926,7 +969,8 @@
 	if(ret == 0 and ferror(stdout)){
 	  int e = errno;
 	  if(debug){
-	    fprintf(stderr, "Error writing encrypted data: %s\n",
+	    fprintf(stderr, "Mandos plugin mandos-client: "
+		    "Error writing encrypted data: %s\n",
 		    strerror(errno));
 	  }
 	  errno = e;
@@ -990,7 +1034,8 @@
   switch(event){
   default:
   case AVAHI_RESOLVER_FAILURE:
-    fprintf(stderr, "(Avahi Resolver) Failed to resolve service '%s'"
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "(Avahi Resolver) Failed to resolve service '%s'"
 	    " of type '%s' in domain '%s': %s\n", name, type, domain,
 	    avahi_strerror(avahi_server_errno(mc.server)));
     break;
@@ -1000,7 +1045,8 @@
       char ip[AVAHI_ADDRESS_STR_MAX];
       avahi_address_snprint(ip, sizeof(ip), address);
       if(debug){
-	fprintf(stderr, "Mandos server \"%s\" found on %s (%s, %"
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Mandos server \"%s\" found on %s (%s, %"
 		PRIdMAX ") on port %" PRIu16 "\n", name, host_name,
 		ip, (intmax_t)interface, port);
       }
@@ -1040,7 +1086,8 @@
   default:
   case AVAHI_BROWSER_FAILURE:
     
-    fprintf(stderr, "(Avahi browser) %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "(Avahi browser) %s\n",
 	    avahi_strerror(avahi_server_errno(mc.server)));
     avahi_simple_poll_quit(mc.simple_poll);
     return;
@@ -1054,7 +1101,8 @@
     if(avahi_s_service_resolver_new(mc.server, interface, protocol,
 				    name, type, domain, protocol, 0,
 				    resolve_callback, NULL) == NULL)
-      fprintf(stderr, "Avahi: Failed to resolve service '%s': %s\n",
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Avahi: Failed to resolve service '%s': %s\n",
 	      name, avahi_strerror(avahi_server_errno(mc.server)));
     break;
     
@@ -1064,7 +1112,8 @@
   case AVAHI_BROWSER_ALL_FOR_NOW:
   case AVAHI_BROWSER_CACHE_EXHAUSTED:
     if(debug){
-      fprintf(stderr, "No Mandos server found, still searching...\n");
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "No Mandos server found, still searching...\n");
     }
     break;
   }
@@ -1109,38 +1158,40 @@
   /* Reject the loopback device */
   if(ifr->ifr_flags & IFF_LOOPBACK){
     if(debug){
-      fprintf(stderr, "Rejecting loopback interface \"%s\"\n",
-	      ifname);
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Rejecting loopback interface \"%s\"\n", ifname);
     }
     return false;
   }
   /* Accept point-to-point devices only if connect_to is specified */
   if(connect_to != NULL and (ifr->ifr_flags & IFF_POINTOPOINT)){
     if(debug){
-      fprintf(stderr, "Accepting point-to-point interface \"%s\"\n",
-	      ifname);
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Accepting point-to-point interface \"%s\"\n", ifname);
     }
     return true;
   }
   /* Otherwise, reject non-broadcast-capable devices */
   if(not (ifr->ifr_flags & IFF_BROADCAST)){
     if(debug){
-      fprintf(stderr, "Rejecting non-broadcast interface \"%s\"\n",
-	      ifname);
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Rejecting non-broadcast interface \"%s\"\n", ifname);
     }
     return false;
   }
   /* Reject non-ARP interfaces (including dummy interfaces) */
   if(ifr->ifr_flags & IFF_NOARP){
     if(debug){
-      fprintf(stderr, "Rejecting non-ARP interface \"%s\"\n", ifname);
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Rejecting non-ARP interface \"%s\"\n", ifname);
     }
     return false;
   }
   
   /* Accept this device */
   if(debug){
-    fprintf(stderr, "Interface \"%s\" is good\n", ifname);
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Interface \"%s\" is good\n", ifname);
   }
   return true;
 }
@@ -1151,13 +1202,63 @@
  * (This function is passed to scandir(3) as a filter function.)
  */
 int good_interface(const struct dirent *if_entry){
-  int ret;
   if(if_entry->d_name[0] == '.'){
     return 0;
   }
+  
   struct ifreq ifr;
+  if(not get_flags(if_entry->d_name, &ifr)){
+    if(debug){
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Failed to get flags for interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  
+  if(not good_flags(if_entry->d_name, &ifr)){
+    return 0;
+  }
+  return 1;
+}
 
+/* 
+ * This function determines if a directory entry in /sys/class/net
+ * corresponds to an acceptable network device which is up.
+ * (This function is passed to scandir(3) as a filter function.)
+ */
+int up_interface(const struct dirent *if_entry){
+  if(if_entry->d_name[0] == '.'){
+    return 0;
+  }
+  
+  struct ifreq ifr;
   if(not get_flags(if_entry->d_name, &ifr)){
+    if(debug){
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Failed to get flags for interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  
+  /* Reject down interfaces */
+  if(not (ifr.ifr_flags & IFF_UP)){
+    if(debug){
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Rejecting down interface \"%s\"\n",
+	      if_entry->d_name);
+    }
+    return 0;
+  }
+  
+  /* Reject non-running interfaces */
+  if(not (ifr.ifr_flags & IFF_RUNNING)){
+    if(debug){
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Rejecting non-running interface \"%s\"\n",
+	      if_entry->d_name);
+    }
     return 0;
   }
   
@@ -1178,6 +1279,58 @@
   return 1;
 }
 
+/* Is this directory entry a runnable program? */
+int runnable_hook(const struct dirent *direntry){
+  int ret;
+  struct stat st;
+  
+  if((direntry->d_name)[0] == '\0'){
+    /* Empty name? */
+    return 0;
+  }
+  
+  /* Save pointer to last character */
+  char *end = strchr(direntry->d_name, '\0')-1;
+  
+  if(*end == '~'){
+    /* Backup name~ */
+    return 0;
+  }
+  
+  if(((direntry->d_name)[0] == '#')
+     and (*end == '#')){
+    /* Temporary #name# */
+    return 0;
+  }
+  
+  /* XXX more rules here */
+  
+  char *fullname = NULL;
+  ret = asprintf(&fullname, "%s/%s", hookdir,
+		 direntry->d_name);
+  if(ret < 0){
+    perror_plus("asprintf");
+    return 0;
+  }
+  
+  ret = stat(fullname, &st);
+  if(ret == -1){
+    if(debug){
+      perror_plus("Could not stat plugin");
+    }
+    return 0;
+  }
+  if(not (S_ISREG(st.st_mode))){
+    /* Not a regular file */
+    return 0;
+  }
+  if(not (st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH))){
+    /* Not executable */
+    return 0;
+  }
+  return 1;
+}
+
 int avahi_loop_with_timeout(AvahiSimplePoll *s, int retry_interval){
   int ret;
   struct timespec now;
@@ -1187,13 +1340,14 @@
   while(true){
     if(mc.current_server == NULL){
       if (debug){
-	fprintf(stderr,
+	fprintf(stderr, "Mandos plugin mandos-client: "
 		"Wait until first server is found. No timeout!\n");
       }
       ret = avahi_simple_poll_iterate(s, -1);
     } else {
       if (debug){
-	fprintf(stderr, "Check current_server if we should run it,"
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Check current_server if we should run it,"
 		" or wait\n");
       }
       /* the current time */
@@ -1216,7 +1370,8 @@
 		    - ((intmax_t)waited_time.tv_nsec / 1000000));
       
       if (debug){
-	fprintf(stderr, "Blocking for %" PRIdMAX " ms\n", block_time);
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Blocking for %" PRIdMAX " ms\n", block_time);
       }
       
       if(block_time <= 0){
@@ -1242,7 +1397,7 @@
       ret = avahi_simple_poll_iterate(s, (int)block_time);
     }
     if(ret != 0){
-      if (ret > 0 or errno != EINTR) {
+      if (ret > 0 or errno != EINTR){
 	return (ret != 1) ? ret : 0;
       }
     }
@@ -1336,6 +1491,10 @@
 	.arg = "SECONDS",
 	.doc = "Retry interval used when denied by the mandos server",
 	.group = 2 },
+      { .name = "network-hook-dir", .key = 133,
+	.arg = "DIR",
+	.doc = "Directory where network hooks are located",
+	.group = 2 },
       /*
        * These reproduce what we would get without ARGP_NO_HELP
        */
@@ -1394,6 +1553,9 @@
 	  argp_error(state, "Bad retry interval");
 	}
 	break;
+      case 133:			/* --network-hook-dir */
+	hookdir = arg;
+	break;
 	/*
 	 * These reproduce what we would get without ARGP_NO_HELP
 	 */
@@ -1405,6 +1567,7 @@
 	argp_state_help(state, state->out_stream,
 			ARGP_HELP_USAGE | ARGP_HELP_EXIT_ERR);
       case 'V':			/* --version */
+	fprintf(state->out_stream, "Mandos plugin mandos-client: ");
 	fprintf(state->out_stream, "%s\n", argp_program_version);
 	exit(argp_err_exit_status);
 	break;
@@ -1495,19 +1658,121 @@
     }
   }
   
+  /* Find network hooks and run them */
+  {
+    struct dirent **direntries;
+    struct dirent *direntry;
+    int numhooks = scandir(hookdir, &direntries, runnable_hook,
+			   alphasort);
+    if(numhooks == -1){
+      perror_plus("scandir");
+    } else {
+      int devnull = open("/dev/null", O_RDONLY);
+      for(int i = 0; i < numhooks; i++){
+	direntry = direntries[0];
+	char *fullname = NULL;
+	ret = asprintf(&fullname, "%s/%s", hookdir, direntry->d_name);
+	if(ret < 0){
+	  perror_plus("asprintf");
+	  continue;
+	}
+	pid_t hook_pid = fork();
+	if(hook_pid == 0){
+	  /* Child */
+	  dup2(devnull, STDIN_FILENO);
+	  close(devnull);
+	  dup2(STDERR_FILENO, STDOUT_FILENO);
+	  ret = setenv("DEVICE", interface, 1);
+	  if(ret == -1){
+	    perror_plus("setenv");
+	    exit(1);
+	  }
+	  ret = setenv("VERBOSE", debug ? "1" : "0", 1);
+	  if(ret == -1){
+	    perror_plus("setenv");
+	    exit(1);
+	  }
+	  ret = setenv("MODE", "start", 1);
+	  if(ret == -1){
+	    perror_plus("setenv");
+	    exit(1);
+	  }
+	  char *delaystring;
+	  ret = asprintf(&delaystring, "%f", delay);
+	  if(ret == -1){
+	    perror_plus("asprintf");
+	    exit(1);
+	  }
+	  ret = setenv("DELAY", delaystring, 1);
+	  if(ret == -1){
+	    free(delaystring);
+	    perror_plus("setenv");
+	    exit(1);
+	  }
+	  free(delaystring);
+	  ret = execl(fullname, direntry->d_name, "start", NULL);
+	  perror_plus("execl");
+	} else {
+	  int status;
+	  if(TEMP_FAILURE_RETRY(waitpid(hook_pid, &status, 0)) == -1){
+	    perror_plus("waitpid");
+	    free(fullname);
+	    continue;
+	  }
+	  if(WIFEXITED(status)){
+	    if(WEXITSTATUS(status) != 0){
+	      fprintf(stderr, "Mandos plugin mandos-client: "
+		      "Warning: network hook \"%s\" exited"
+		      " with status %d\n", direntry->d_name,
+		      WEXITSTATUS(status));
+	      free(fullname);
+	      continue;
+	    }
+	  } else if(WIFSIGNALED(status)){
+	    fprintf(stderr, "Mandos plugin mandos-client: "
+		    "Warning: network hook \"%s\" died by"
+		    " signal %d\n", direntry->d_name,
+		    WTERMSIG(status));
+	    free(fullname);
+	    continue;
+	  } else {
+	    fprintf(stderr, "Mandos plugin mandos-client: "
+		    "Warning: network hook \"%s\" crashed\n",
+		    direntry->d_name);
+	    free(fullname);
+	    continue;
+	  }
+	}
+	free(fullname);
+	if(quit_now){
+	  goto end;
+	}
+      }
+      close(devnull);
+    }
+  }
+  
   if(not debug){
     avahi_set_log_function(empty_log);
   }
-
+  
   if(interface[0] == '\0'){
     struct dirent **direntries;
-    ret = scandir(sys_class_net, &direntries, good_interface,
+    /* First look for interfaces that are up */
+    ret = scandir(sys_class_net, &direntries, up_interface,
 		  alphasort);
+    if(ret == 0){
+      /* No up interfaces, look for any good interfaces */
+      free(direntries);
+      ret = scandir(sys_class_net, &direntries, good_interface,
+		    alphasort);
+    }
     if(ret >= 1){
-      /* Pick the first good interface */
+      /* Pick the first interface returned */
       interface = strdup(direntries[0]->d_name);
       if(debug){
-	fprintf(stderr, "Using interface \"%s\"\n", interface);
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Using interface \"%s\"\n", interface);
       }
       if(interface == NULL){
 	perror_plus("malloc");
@@ -1518,7 +1783,8 @@
       free(direntries);
     } else {
       free(direntries);
-      fprintf(stderr, "Could not find a network interface\n");
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Could not find a network interface\n");
       exitcode = EXIT_FAILURE;
       goto end;
     }
@@ -1530,7 +1796,8 @@
   srand((unsigned int) time(NULL));
   mc.simple_poll = avahi_simple_poll_new();
   if(mc.simple_poll == NULL){
-    fprintf(stderr, "Avahi: Failed to create simple poll object.\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Avahi: Failed to create simple poll object.\n");
     exitcode = EX_UNAVAILABLE;
     goto end;
   }
@@ -1602,7 +1869,8 @@
   if(strcmp(interface, "none") != 0){
     if_index = (AvahiIfIndex) if_nametoindex(interface);
     if(if_index == 0){
-      fprintf(stderr, "No such interface: \"%s\"\n", interface);
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "No such interface: \"%s\"\n", interface);
       exitcode = EX_UNAVAILABLE;
       goto end;
     }
@@ -1749,7 +2017,8 @@
   
   ret = init_gnutls_global(pubkey, seckey);
   if(ret == -1){
-    fprintf(stderr, "init_gnutls_global failed\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "init_gnutls_global failed\n");
     exitcode = EX_UNAVAILABLE;
     goto end;
   } else {
@@ -1771,7 +2040,8 @@
   }
   
   if(not init_gpgme(pubkey, seckey, tempdir)){
-    fprintf(stderr, "init_gpgme failed\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "init_gpgme failed\n");
     exitcode = EX_UNAVAILABLE;
     goto end;
   } else {
@@ -1787,7 +2057,8 @@
     /* (Mainly meant for debugging) */
     char *address = strrchr(connect_to, ':');
     if(address == NULL){
-      fprintf(stderr, "No colon in address\n");
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "No colon in address\n");
       exitcode = EX_USAGE;
       goto end;
     }
@@ -1801,7 +2072,8 @@
     tmpmax = strtoimax(address+1, &tmp, 10);
     if(errno != 0 or tmp == address+1 or *tmp != '\0'
        or tmpmax != (uint16_t)tmpmax){
-      fprintf(stderr, "Bad port number\n");
+      fprintf(stderr, "Mandos plugin mandos-client: "
+	      "Bad port number\n");
       exitcode = EX_USAGE;
       goto end;
     }
@@ -1837,8 +2109,8 @@
 	break;
       }
       if(debug){
-	fprintf(stderr, "Retrying in %d seconds\n",
-		(int)retry_interval);
+	fprintf(stderr, "Mandos plugin mandos-client: "
+		"Retrying in %d seconds\n", (int)retry_interval);
       }
       sleep((int)retry_interval);
     }
@@ -1846,7 +2118,7 @@
     if (not quit_now){
       exitcode = EXIT_SUCCESS;
     }
-
+    
     goto end;
   }
   
@@ -1874,7 +2146,8 @@
   
   /* Check if creating the Avahi server object succeeded */
   if(mc.server == NULL){
-    fprintf(stderr, "Failed to create Avahi server: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Failed to create Avahi server: %s\n",
 	    avahi_strerror(error));
     exitcode = EX_UNAVAILABLE;
     goto end;
@@ -1889,7 +2162,8 @@
 				   AVAHI_PROTO_UNSPEC, "_mandos._tcp",
 				   NULL, 0, browse_callback, NULL);
   if(sb == NULL){
-    fprintf(stderr, "Failed to create service browser: %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Failed to create service browser: %s\n",
 	    avahi_strerror(avahi_server_errno(mc.server)));
     exitcode = EX_UNAVAILABLE;
     goto end;
@@ -1902,20 +2176,23 @@
   /* Run the main loop */
   
   if(debug){
-    fprintf(stderr, "Starting Avahi loop search\n");
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "Starting Avahi loop search\n");
   }
 
   ret = avahi_loop_with_timeout(mc.simple_poll,
 				(int)(retry_interval * 1000));
   if(debug){
-    fprintf(stderr, "avahi_loop_with_timeout exited %s\n",
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "avahi_loop_with_timeout exited %s\n",
 	    (ret == 0) ? "successfully" : "with error");
   }
   
  end:
   
   if(debug){
-    fprintf(stderr, "%s exiting\n", argv[0]);
+    fprintf(stderr, "Mandos plugin mandos-client: "
+	    "%s exiting\n", argv[0]);
   }
   
   /* Cleanup things */
@@ -1949,6 +2226,8 @@
     }
   }
   
+  /* XXX run network hooks "stop" here  */
+  
   /* Take down the network interface */
   if(take_down_interface){
     /* Re-raise priviliges */
@@ -1961,7 +2240,7 @@
       ret = ioctl(sd, SIOCGIFFLAGS, &network);
       if(ret == -1){
 	perror_plus("ioctl SIOCGIFFLAGS");
-      } else if(network.ifr_flags & IFF_UP) {
+      } else if(network.ifr_flags & IFF_UP){
 	network.ifr_flags &= ~(short)IFF_UP; /* clear flag */
 	ret = ioctl(sd, SIOCSIFFLAGS, &network);
 	if(ret == -1){
@@ -1999,8 +2278,8 @@
 	}
 	ret = remove(fullname);
 	if(ret == -1){
-	  fprintf(stderr, "remove(\"%s\"): %s\n", fullname,
-		  strerror(errno));
+	  fprintf(stderr, "Mandos plugin mandos-client: "
+		  "remove(\"%s\"): %s\n", fullname, strerror(errno));
 	}
 	free(fullname);
       }

=== modified file 'plugins.d/mandos-client.xml'
--- plugins.d/mandos-client.xml	2011-10-05 16:00:56 +0000
+++ plugins.d/mandos-client.xml	2011-11-12 18:04:35 +0000
@@ -102,6 +102,10 @@
       </arg>
       <sbr/>
       <arg>
+	<option>--network-hook-dir<replaceable>DIR</replaceable></option>
+      </arg>
+      <sbr/>
+      <arg>
 	<option>--debug</option>
       </arg>
     </cmdsynopsis>
@@ -311,6 +315,18 @@
 	  </para>
 	</listitem>
       </varlistentry>
+
+      <varlistentry>
+	<term><option>--network-hook-dir=<replaceable
+	>DIR</replaceable></option></term>
+	<listitem>
+	  <para>
+	    Network hook directory.  The default directory is
+	    <quote><directory class="directory"
+	    >/lib/mandos/network-hooks.d</directory></quote>.
+	  </para>
+	</listitem>
+      </varlistentry>
       
       <varlistentry>
 	<term><option>--debug</option></term>

